@@ -1,14 +1,25 @@
 ##=== JDBCDriver
 
-setClass("JDBCDriver", representation("DBIDriver", identifier.quote="character"))
+setClass("JDBCDriver", representation("DBIDriver", identifier.quote="character", jdrv="jobjRef"))
 
 JDBC <- function(driverClass='', classPath='', identifier.quote=NA) {
   .jinit(classPath)
-  if (nchar(driverClass))
-    if (is.jnull(.jcall("java/lang/Class","Ljava/lang/Class;", "forName", as.character(driverClass)[1],
-                        TRUE, .jcall("java/lang/ClassLoader", "Ljava/lang/ClassLoader;", "getSystemClassLoader"))))
-      stop("Cannot find JDBC driver class ",driverClass)
-  new("JDBCDriver", identifier.quote=as.character(identifier.quote))
+  if (nchar(driverClass) && is.jnull(.jfindClass(as.character(driverClass)[1])))
+    stop("Cannot find JDBC driver class ",driverClass)
+  jdrv <- .jnew(driverClass, check=FALSE)
+  .jcheck(TRUE)
+  if (is.jnull(jdrv)) jdrv <- .jnull()
+  new("JDBCDriver", identifier.quote=as.character(identifier.quote), jdrv=jdrv)
+}
+
+.verify.JDBC.result <- function (result, ...) {
+  if (is.jnull(result)) {
+    x <- .jgetEx(TRUE)
+    if (is.jnull(x))
+      stop(...)
+    else
+      stop(...," (",.jcall(x, "S", "getMessage"),")")
+  }
 }
 
 setMethod("dbListConnections", "JDBCDriver", def=function(drv, ...) { warning("JDBC driver maintains no list of acitve connections."); list() })
@@ -23,8 +34,17 @@ setMethod("dbGetInfo", "JDBCDriver", def=function(dbObj, ...)
 setMethod("dbUnloadDriver", "JDBCDriver", def=function(drv, ...) NULL)
 
 setMethod("dbConnect", "JDBCDriver", def=function(drv, url, user='', password='', ...) {
-  jc <- .jcall("java/sql/DriverManager","Ljava/sql/Connection;","getConnection", as.character(url)[1], as.character(user)[1], as.character(password)[1])
-  if (is.jnull(jc)) stop("Unable to connect JDBC to ",url)
+  jc <- .jcall("java/sql/DriverManager","Ljava/sql/Connection;","getConnection", as.character(url)[1], as.character(user)[1], as.character(password)[1], check=FALSE)
+  if (is.jnull(jc) || !is.jnull(drv@jdrv)) {
+    # ok one reason for this to fail is its interaction with rJava's
+    # class loader. In that case we try to load the driver directly.
+    oex <- .jgetEx(TRUE)
+    p <- .jnew("java/util/Properties")
+    if (length(user)==1 && nchar(user)) .jcall(p,"Ljava/lang/Object;","setProperty","user",user)
+    if (length(password)==1 && nchar(password)) .jcall(p,"Ljava/lang/Object;","setProperty","password",password)
+    jc <- .jcall(drv@jdrv, "Ljava/sql/Connection;", "connect", as.character(url)[1], p)
+  }
+  .verify.JDBC.result(jc, "Unable to connect JDBC to ",url)
   new("JDBCConnection", jc=jc, identifier.quote=drv@identifier.quote)},
           valueClass="JDBCConnection")
 
@@ -38,7 +58,7 @@ setMethod("dbDisconnect", "JDBCConnection", def=function(conn, ...)
 .fillStatementParameters <- function(s, l) {
   for (i in 1:length(l)) {
     if (is.integer(l[[i]]))
-      .jcall(s, "V", "setInteger", i, l[[i]][1])
+      .jcall(s, "V", "setInt", i, l[[i]][1])
     else {
       if (is.numeric(l[[i]]))
         .jcall(s, "V", "setDouble", i, as.double(l[[i]])[1])
@@ -49,26 +69,27 @@ setMethod("dbDisconnect", "JDBCConnection", def=function(conn, ...)
 }
 
 setMethod("dbSendQuery", signature(conn="JDBCConnection", statement="character"),  def=function(conn, statement, ..., list=NULL) {
-  s <- .jcall(conn@jc, "Ljava/sql/PreparedStatement;", "prepareStatement", as.character(statement)[1])
-  if (is.jnull(s)) stop("Unable to execute JDBC statement ",statement)
+  s <- .jcall(conn@jc, "Ljava/sql/PreparedStatement;", "prepareStatement", as.character(statement)[1], check=FALSE)
+  .verify.JDBC.result(s, "Unable to execute JDBC statement ",statement)
   if (length(list(...))) .fillStatementParameters(s, list(...))
   if (!is.null(list)) .fillStatementParameters(s, list)
-  r <- .jcall(s, "Ljava/sql/ResultSet;", "executeQuery")
-  if (is.jnull(r)) stop("Unable to retrieve JDBC result set for ",statement)
-  md <- .jcall(r, "Ljava/sql/ResultSetMetaData;", "getMetaData")
-  if (is.jnull(md)) stop("Unable to retrieve JDBC result set meta data for ",statement, " in dbSendQuery")
+  r <- .jcall(s, "Ljava/sql/ResultSet;", "executeQuery", check=FALSE)
+  .verify.JDBC.result(r, "Unable to retrieve JDBC result set for ",statement)
+  md <- .jcall(r, "Ljava/sql/ResultSetMetaData;", "getMetaData", check=FALSE)
+  .verify.JDBC.result(md, "Unable to retrieve JDBC result set meta data for ",statement, " in dbSendQuery")
   new("JDBCResult", jr=r, md=md)
 })
 
 if (is.null(getGeneric("dbSendUpdate"))) setGeneric("dbSendUpdate", function(conn, statement, ...) standardGeneric("dbSendUpdate"))
 
 setMethod("dbSendUpdate",  signature(conn="JDBCConnection", statement="character"),  def=function(conn, statement, ..., list=NULL) {
-  s <- .jcall(conn@jc, "Ljava/sql/PreparedStatement;", "prepareStatement", as.character(statement)[1])
-  if (is.jnull(s)) stop("Unable to execute JDBC statement ",statement)
+  s <- .jcall(conn@jc, "Ljava/sql/PreparedStatement;", "prepareStatement", as.character(statement)[1], check=FALSE)
+  .verify.JDBC.result(s, "Unable to execute JDBC statement ",statement)
   if (length(list(...))) .fillStatementParameters(s, list(...))
   if (!is.null(list)) .fillStatementParameters(s, list)
   .jcall(s, "I", "executeUpdate", check=FALSE)
-  if (.jcheck()) stop("execute JDBC update query failed in dbSendUpdate")
+  x <- .jgetEx(TRUE)
+  if (!is.jnull(x)) stop("execute JDBC update query failed in dbSendUpdate (", .jcall(x, "S", "getMessage"),")")
 })
 
 setMethod("dbGetQuery", signature(conn="JDBCConnection", statement="character"),  def=function(conn, statement, ...) {
@@ -88,11 +109,11 @@ setMethod("dbListResults", "JDBCConnection",
           )
 
 setMethod("dbListTables", "JDBCConnection", def=function(conn, pattern="%", ...) {
-  md <- .jcall(conn@jc, "Ljava/sql/DatabaseMetaData;", "getMetaData")
-  if (is.jnull(md)) stop("Unable to retrieve JDBC database metadata")
+  md <- .jcall(conn@jc, "Ljava/sql/DatabaseMetaData;", "getMetaData", check=FALSE)
+  .verify.JDBC.result(md, "Unable to retrieve JDBC database metadata")
   r <- .jcall(md, "Ljava/sql/ResultSet;", "getTables", .jnull("java/lang/String"),
-              .jnull("java/lang/String"), pattern, .jnull("[Ljava/lang/String;"))
-  if (is.jnull(r)) stop("Unable to retrieve JDBC tables list")
+              .jnull("java/lang/String"), pattern, .jnull("[Ljava/lang/String;"), check=FALSE)
+  .verify.JDBC.result(r, "Unable to retrieve JDBC tables list")
   ts <- character()
   while (.jcall(r, "Z", "next"))
     ts <- c(ts, .jcall(r, "S", "getString", "TABLE_NAME"))
@@ -105,11 +126,11 @@ setMethod("dbExistsTable", "JDBCConnection", def=function(conn, name, ...) (leng
 setMethod("dbRemoveTable", "JDBCConnection", def=function(conn, name, ...) dbSendUpdate(conn, paste("DROP TABLE", name))==0)
 
 setMethod("dbListFields", "JDBCConnection", def=function(conn, name, pattern="%", ...) {
-  md <- .jcall(conn@jc, "Ljava/sql/DatabaseMetaData;", "getMetaData")
-  if (is.jnull(md)) stop("Unable to retrieve JDBC database metadata")
+  md <- .jcall(conn@jc, "Ljava/sql/DatabaseMetaData;", "getMetaData", check=FALSE)
+  .verify.JDBC.result(md, "Unable to retrieve JDBC database metadata")
   r <- .jcall(md, "Ljava/sql/ResultSet;", "getColumns", .jnull("java/lang/String"),
-              .jnull("java/lang/String"), name, pattern)
-  if (is.jnull(r)) stop("Unable to retrieve JDBC columns list for ",name)
+              .jnull("java/lang/String"), name, pattern, check=FALSE)
+  .verify.JDBC.result(r, "Unable to retrieve JDBC columns list for ",name)
   ts <- character()
   while (.jcall(r, "Z", "next"))
     ts <- c(ts, .jcall(r, "S", "getString", "COLUMN_NAME"))
@@ -160,14 +181,14 @@ setMethod("dbWriteTable", "JDBCConnection", def=function(conn, name, value, over
   }
   fdef <- paste(.sql.qescape(names(value), TRUE, conn@identifier.quote),fts,collapse=',')
   qname <- .sql.qescape(name, TRUE, conn@identifier.quote)
-  ct <- paste("CREATE TABLE ",qname," (",fdef,");",sep= '')
+  ct <- paste("CREATE TABLE ",qname," (",fdef,")",sep= '')
   if (ac) {
     .jcall(conn@jc, "V", "setAutoCommit", FALSE)
     on.exit(.jcall(conn@jc, "V", "setAutoCommit", ac))
   }
   dbSendUpdate(conn, ct)
   if (length(value[[1]])) {
-    inss <- paste("INSERT INTO ",qname," VALUES(", paste(rep("?",length(value)),collapse=','),");",sep='')
+    inss <- paste("INSERT INTO ",qname," VALUES(", paste(rep("?",length(value)),collapse=','),")",sep='')
     for (j in 1:length(value[[1]]))
       dbSendUpdate(conn, inss, list=as.list(value[j,]))
   }
